// validating resolver

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	recursion yes;
	dnssec-validation yes;
};

controls {
	inet 10.53.0.2 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

include "../../_common/rndc.key";

zone "." {
	type hint;
	file "../../_common/root.hint";
};

zone "tld.test" {
	type static-stub;
	server-addresses { 10.53.0.1; };
};

trust-anchors {
	tld.test. static-key 257 3 13 "@TLD_DNSKEY@";
};
