options {
	query-source address 10.53.0.1;
	notify-source 10.53.0.1;
	transfer-source 10.53.0.1;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.1; };
	listen-on-v6 { none; };
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.1 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

view chaos ch {
	match-clients { any; };
	recursion yes;
	zone "." {
		type hint;
		file "chaos.db";
	};
	zone "version.bind" {
		type primary;
		database "_builtin version";
	};
};
