/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

{% set server_config_use_addr = server_config_use_addr | default(False) %}

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.2 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

include "trusted.conf";

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	recursion yes;
	dnssec-validation yes;
	notify no;
	minimal-responses no;
};

zone "." {
	type hint;
	file "../../_common/root.hint";
};

zone "example" {
	type static-stub;
	server-addresses { 10.53.0.3; };
	allow-query { !10.53.0.7; any; };
};

zone "example.org" {
	type static-stub;
{% if server_config_use_addr %}
	server-addresses { 10.53.0.4; };
{% else %}
	server-names { "ns.example.net"; };
{% endif %}
};

zone "example.info" {
	type static-stub;
	server-addresses { ::1; }; #ns4
};

zone "undelegated" {
	type static-stub;
	server-addresses { 10.53.0.3; };
};

zone "unsigned" {
	type static-stub;
	server-addresses { 10.53.0.3; };
};
