options {
    query-source address @ns.ip@;
    port @PORT@;
    pid-file "named.pid";
    listen-on { @ns.ip@; };
    listen-on-v6 { none; };
    recursion yes;
    dnssec-validation no;
};

{% include "_common/root.hint.conf" %}
