options {
	directory ".";
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.2 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

view default {
	match-clients { any; };
	recursion no;
        dnssec-validation no;
	zone "1.0.0.127.in-addr.arpa." {
		type primary;
		file "localhost.db";
		update-policy {
			grant * tcp-self . ANY;
		};
	};
};

view chaos ch {
	match-clients { any; };
	recursion no;
	zone example {
		type primary;
		file "example.db";
		allow-update { any; };
	};
};
