// NS7

options {
	port @PORT@;
	listen-on port @PORT@ { 10.53.0.7; };
	pid-file "named.pid";
	nxdomain-redirect redirect;
	dnssec-validation no;
	dns64 64:ff9b::/96 {
		clients { any; };
		mapped { any; };
		suffix ::;
	};
};

zone "." {
        type primary;
        file "root.db";
};

zone "redirect" {
        type primary;
        file "redirect.db";
};
