options {
	directory ".";
	query-source address 10.53.0.3;
	notify-source 10.53.0.3;
	transfer-source 10.53.0.3;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.3; };
	listen-on-v6 { none; };
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.3 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

view chaos ch {
	match-clients { any; };
	recursion yes;
        dnssec-validation no;
        forward only;
        forwarders port @PORT@ { 10.53.0.2; };
        deny-answer-addresses { 0.0.0.0/0; ::/0; };
};
